@@ -22,9 +22,13 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 
 '''
-import logging, os, re
+import logging
+import os
+import re
+
 import colorama
 
+
 class NullIndenter(object):
     def __enter__(self):
         return self
