@@ -1,5 +1,5 @@
 '''
-BLURB GOES HERE.
+Maestro is my crappy, synchronous, yet extensible build system
 
 Copyright (c) 2015 - 2022 Rob "N3X15" Nelson <nexisentertainment@gmail.com>
 
@@ -29,20 +29,16 @@ import re
 import sys
 import shutil
 import argparse
+import traceback
 
-from collections import defaultdict
 from buildtools import os_utils
-from buildtools.bt_logging import NullIndenter, log
+from buildtools.bt_logging import log
 from buildtools.maestro.base_target import BuildTarget
-from buildtools.maestro.fileio import (ConcatenateBuildTarget, CopyFilesTarget,
-                                       CopyFileTarget, MoveFileTarget,
-                                       ReplaceTextTarget)
 from buildtools.maestro.utils import (SerializableFileLambda,
-                                      SerializableLambda, callLambda)
+                                      SerializableLambda)
 
 from ruamel.yaml import YAML
-from typing import List
-from tqdm import tqdm
+from typing import Dict, List
 
 yaml = YAML(typ='safe', pure=True)
 
@@ -51,7 +47,7 @@ yaml.register_class(SerializableFileLambda)
 
 class TarjanGraphVertex(object):
     def __init__(self, ID: int, refs: List[int]):
-        self.ID=ID
+        self.id=ID
         self.refs = refs
         self.disc = -1
         self.low = -1
@@ -75,7 +71,7 @@ class TarjanGraph(object):
         vertex.disc = self.time
         vertex.low = self.time
         self.time+=1
-        stack.append(vertex.ID)
+        stack.append(vertex.id)
         vertex.stackMember=True
 
         for vID in vertex.refs:
@@ -88,7 +84,7 @@ class TarjanGraph(object):
         w=-1
         if vertex.low == vertex.disc:
             cycle=[]
-            while w!=vertex.ID:
+            while w!=vertex.id:
                 w = stack.pop()
                 cycle += [w]
                 self.vertices[w].stackMember=False
@@ -100,32 +96,51 @@ class TarjanGraph(object):
             if vertex.disc == -1:
                 self._sccutil(vertex, stack)
 
-class BuildMaestro(object):
+class BuildMaestro:
     ALL_TYPES = {}
 
     def __init__(self, hidden_build_dir='.build'):
-        self.alltargets = []
-        self.targets = []
-        self.targetsCompleted = []
+        self.alltargets: List[BuildTarget] = []
+        self.targets: List[str] = []
+        self.targetsCompleted: List[str] = []
+        self.targetsProvidedBy: Dict[str, BuildTarget] = {}
 
-        self.verbose = False
-        self.colors = False
-        self.show_commands = False
+        self.verbose: bool = False
+        self.colors: bool = False
+        self.show_commands: bool = False
 
-        self.builddir = hidden_build_dir
-        self.all_targets_file = os.path.join(self.builddir, 'all_targets.yml')
+        self.builddir: str = hidden_build_dir
+        self.all_targets_file: str = os.path.join(self.builddir, 'all_targets.yml')
 
         # This will get Maestro to delete the listed directories, if they are present during --clean.
         self.other_dirs_to_clean=[]
 
         self.args: argparse.Namespace = None
 
-    def add(self, bt):
-        bt.ID = len(self.alltargets)
+    def add(self, bt: BuildTarget) -> BuildTarget:
+        bt.added_at = traceback.extract_stack()
+        bt.id = len(self.alltargets)
         self.alltargets.append(bt)
-        self.targets += bt.provides()
+        for t in bt.provides():
+            if t in self.targets:
+                log.warn(f'Target {t} already exists.')
+                old = self.targetsProvidedBy[t]
+                with log.warn('Old:'):
+                    old._dumpStacks()
+                with log.warn('New:'):
+                    bt._dumpStacks()
+            self.targetsProvidedBy[t] = bt
+            self.targets.append(t)
         return bt
 
+    def generateVirtualTarget(self, id: str) -> str:
+        '''
+        Usage:
+        
+        bm.add(CopyFilesTarget(target=bm.generateVirtualTarget('a'),...))
+        '''
+        return os.path.join(self.builddir, '.targets', id)
+
     def build_argparser(self):
         argp = argparse.ArgumentParser()
         argp.add_argument('--clean', action='store_true', default=False, help='Cleans everything.')
@@ -220,60 +235,59 @@ class BuildMaestro(object):
                     yaml.dump(target, f)
                 f.write(u'\n')
 
-    def loadRules(self, filename):
-        REGEX_RULEHEADER = re.compile('\[([A-Za-z0-9]+) ([^:]+)\]:(.*)$')
-        self.targets = []
-        self.alltargets = []
-        with codecs.open(filename, 'r') as f:
-            context = {}
-            yamlbuf = ''
-            ruleKey = ''
-            for oline in f:
-                s_line = oline.strip()
-                if s_line.startswith('#') or s_line == '':
-                    continue
-                line = oline.rstrip()
-                m = REGEX_RULEHEADER.match(line)
-                if m is not None:
-                    if len(context.keys()) > 0:
-                        self.addFromRules(context, yamlbuf)
-                        context = None
-                        yamlbuf = ''
-                        ruleKey = ''
-                    typeID, ruleKey, depends = m.group(1, 2, 3)
-                    context = {
-                        'type': typeID,
-                        'target': ruleKey,
-                        'dependencies': [x.strip() for x in depends.split(',') if x != ''],
-                        'files': [],
-                        'provides': []
-                    }
-                elif line.startswith('>') and context is not None:
-                    context['files'].append(line[1:].strip())
-                elif line.startswith('<') and context is not None:
-                    context['provides'].append(line[1:].strip())
-                else:
-                    yamlbuf += oline
-            if context is not None:
-                self.addFromRules(context, yamlbuf)
-        log.info('Loaded %d rules from %s', len(self.alltargets), filename)
+    # def loadRules(self, filename):
+    #     REGEX_RULEHEADER = re.compile('\[([A-Za-z0-9]+) ([^:]+)\]:(.*)$')
+    #     self.targets = []
+    #     self.alltargets = []
+    #     with codecs.open(filename, 'r') as f:
+    #         context = {}
+    #         yamlbuf = ''
+    #         ruleKey = ''
+    #         for oline in f:
+    #             s_line = oline.strip()
+    #             if s_line.startswith('#') or s_line == '':
+    #                 continue
+    #             line = oline.rstrip()
+    #             m = REGEX_RULEHEADER.match(line)
+    #             if m is not None:
+    #                 if len(context.keys()) > 0:
+    #                     self.addFromRules(context, yamlbuf)
+    #                     context = None
+    #                     yamlbuf = ''
+    #                     ruleKey = ''
+    #                 typeID, ruleKey, depends = m.group(1, 2, 3)
+    #                 context = {
+    #                     'type': typeID,
+    #                     'target': ruleKey,
+    #                     'dependencies': [x.strip() for x in depends.split(',') if x != ''],
+    #                     'files': [],
+    #                     'provides': []
+    #                 }
+    #             elif line.startswith('>') and context is not None:
+    #                 context['files'].append(line[1:].strip())
+    #             elif line.startswith('<') and context is not None:
+    #                 context['provides'].append(line[1:].strip())
+    #             else:
+    #                 yamlbuf += oline
+    #         if context is not None:
+    #             self.addFromRules(context, yamlbuf)
+    #     log.info('Loaded %d rules from %s', len(self.alltargets), filename)
 
-    def addFromRules(self, context, yamlbuf):
-        # print(repr(yamlbuf))
-        if yamlbuf.strip() != '':
-            yml = yaml.full_load(yamlbuf)
-            for k, v in yml.items():
-                context[k] = v
-        cls = self.ALL_TYPES[context['type']]
-        bt = cls()
-        bt.deserialize(context)
-        self.add(bt)
+    # def addFromRules(self, context, yamlbuf):
+    #     # print(repr(yamlbuf))
+    #     if yamlbuf.strip() != '':
+    #         yml = yaml.full_load(yamlbuf)
+    #         for k, v in yml.items():
+    #             context[k] = v
+    #     cls = self.ALL_TYPES[context['type']]
+    #     bt = cls()
+    #     bt.deserialize(context)
+    #     self.add(bt)
 
-    def get_max_label_length(self):
-        max_len = 0
-        for bt in self.alltargets:
-            max_len = max(max_len, len(bt.get_label()))
-        return max_len
+    def get_max_label_length(self) -> int:
+        if len(self.alltargets):
+            return max([len(bt.get_label()) for bt in self.alltargets])
+        return 0
 
     def checkForCycles(self):
         with log.info('Checking for dependency cycles...'):
@@ -297,12 +311,12 @@ class BuildMaestro(object):
                     elif len(providers) == 0:
                         log.critical('Build target %s has no providers for dependency %s: %r',bt.name,depend,[x.name for x in providers])
                         sys.exit(1)
-                    refs.append(providers[-1].ID)
+                    refs.append(providers[-1].id)
                 with log.debug('Dependency tree:'):
-                    with log.debug('[%s] (%d,[%s])',bt.name,bt.ID,', '.join([str(x) for x in refs])):
+                    with log.debug('[%s] (%d,[%s])',bt.name,bt.id,', '.join([str(x) for x in refs])):
                         for refID in refs:
                             log.debug(self.alltargets[refID].name)
-                tg.add_edge(bt.ID, refs)
+                tg.add_edge(bt.id, refs)
 
             # Run the algo
             tg.SCC()
@@ -311,7 +325,7 @@ class BuildMaestro(object):
             foundCycles=False
             for cycle in tg.cycles:
                 if len(cycle) > 1:
-                    log.critical('CYCLE FOUND: %r', ['#{} ({})'.format(self.alltargets[btid].ID,self.alltargets[btid].name) for btid in cycle])
+                    log.critical('CYCLE FOUND: %r', ['#{} ({})'.format(self.alltargets[btid].id,self.alltargets[btid].name) for btid in cycle])
                     foundCycles=True
             return foundCycles
 
